# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for building and applying various specification.

ExclusionSpecification holds an expression is checked against a particular
GaarfRow object to verify whether or not this expression is true.
"""

from __future__ import annotations

import dataclasses
import itertools
import logging
import math
import re
from collections.abc import Sequence

from gaarf import report

from googleads_housekeeper.domain.core import rules_parser
from googleads_housekeeper.services import enums

_PARSEABLE_PLACEMENT_TYPES = ('WEBSITE', 'YOUTUBE_VIDEO', 'YOUTUBE_CHANNEL')
_PARSEABLE_PLACEMENT_TYPES_ENUMS = {
  enums.ExclusionTypeEnum.WEBSITE_INFO,
  enums.ExclusionTypeEnum.YOUTUBE_VIDEO_INFO,
  enums.ExclusionTypeEnum.YOUTUBE_CHANNEL_INFO,
}


@dataclasses.dataclass(frozen=True)
class RuntimeOptions:
  """Defines options for applying specifications.

  Attributes:
    is_conversion_query: Whether Gaarf query contains conversion split.
    conversion_name: Name(s) of conversion(s) for conversion query.
    conversion_rules: All rules related to conversion split.
  """

  is_conversion_query: bool = False
  conversion_name: str = ''
  conversion_rules: list[BaseExclusionSpecificationEntry] = dataclasses.field(
    default_factory=list
  )


class BaseExclusionSpecificationEntry:
  """Base class for holding logic applicable to all specifications.

  Attributes:
    name:
      Name of the metric/dimension from Google Ads.
    operator:
      Comparator used to evaluate the expression.
    value:
      Expected value metric should take.
    exclusion_type:
      Type of the parseable entity, depending on it additional data might
      be fetched from the repository.
  """

  def __init__(
    self, expression: str, exclusion_type: enums.ExclusionTypeEnum
  ) -> None:
    """Constructor for the class.

    Args:
      expression: Exclusion expression in a form of `name > value`.
      exclusion_type: Type of the exclusion.
    """
    elements = [
      element.strip() for element in expression.split(' ', maxsplit=2)
    ]
    if len(elements) != 3:
      raise ValueError("Incorrect expression, specify in 'name > value' format")
    if elements[1] not in (
      '>',
      '>=',
      '<',
      '<=',
      '=',
      '!=',
      'regexp',
      'contains',
    ):
      raise ValueError(
        'Incorrect operator for expression, '
        "only '>', '>=', '<', '<=', '=', '!=', 'regexp', 'contains' ",
      )

    self.name = elements[0]
    self.operator = '==' if elements[1] == '=' else elements[1]
    self.__raw_value = elements[2].replace("'", '').replace('"', '')
    if self.__raw_value.lower() == 'true':
        self.value = True
    elif self.__raw_value.lower() == 'false':
        self.value = False
    else:
      self.value = self.__raw_value
    self.exclusion_type = exclusion_type

  @property
  def corresponding_placement_type(self) -> enums.PlacementTypeEnum | None:
    if self.exclusion_type == enums.ExclusionTypeEnum.GOOGLE_ADS_INFO:
      return None
    return enums.PlacementTypeEnum[
      self.exclusion_type.name.replace('_INFO', '')
    ]

  @property
  def repository_name(self) -> str | None:
    if self.exclusion_type == enums.ExclusionTypeEnum.GOOGLE_ADS_INFO:
      return None
    return self.exclusion_type.name.lower()

  def is_satisfied_by(
    self, placement_info: report.GaarfRow
  ) -> tuple[bool, dict]:
    """Verifies whether given entity satisfies stored expression.

    Args:
      placement_info: GaarfRow object that contains entity data.

    Returns:
      Tuple with results of evaluation and all necessary information on
      placement (formatted as a dict).
    """
    placement_as_dict = {}
    if (
      placement_info.placement_type in _PARSEABLE_PLACEMENT_TYPES
      and self.exclusion_type != enums.ExclusionTypeEnum.GOOGLE_ADS_INFO
    ):
      placement = placement_info.extra_info.get(
        self.exclusion_type.name.lower(), {}
      )
      if not placement:
        return False, {}
      placement_as_dict = dataclasses.asdict(placement)
    else:
      placement = placement_info
    if not placement_as_dict:
      placement_as_dict = placement.to_dict()
    if not hasattr(placement, self.name):
      raise ValueError(f'{placement} has no {self.name} attribute!')
    if hasattr(placement, 'is_processed') and not placement.is_processed:
      logging.debug(
        'Cannot get internal information on %s placement of type %s',
        placement_info.placement,
        placement_info.placement_type,
      )
      return False, {}
    if self.operator in ('regexp', 'contains'):
      result = self._check_regexp(placement)
      return result, {} if not result else placement_as_dict
    result = self._eval_expression(placement)
    return result, {} if not result else placement_as_dict

  def _check_regexp(self, placement: report.GaarfRow) -> bool:
    if placement_element := getattr(placement, self.name):
      return bool(
        re.search(
          rf'{self.value}',
          re.sub(r'[,.;@#?!&$]+', '', placement_element),
          re.IGNORECASE,
        )
      )
    return False

  def _eval_expression(self, placement):
    try:
      value = float(self.value)
    except ValueError:
      value = self.value
    if isinstance(value, float):
      return eval(
        f'{self._nan_to_zero(getattr(placement, self.name))}{self.operator} {value}'
      )
    return getattr(placement, self.name) == value

  def _nan_to_zero(self, value: str) -> float | str:
    return 0.0 if math.isnan(value) else value

  def __str__(self):
    return (
      f'{self.exclusion_type.name}:{self.name} {self.operator} {self.value}'
    )

  def __repr__(self):
    return (
      f'{self.__class__.__name__}'
      f"(exclusion_type='{self.exclusion_type.name}', "
      f"name='{self.name}', "
      f"operator='{self.operator}', value='{self.value}')"
    )

  def __eq__(self, other):
    return (self.exclusion_type, self.name, self.operator, self.value) == (
      other.exclusion_type,
      other.name,
      other.operator,
      other.value,
    )


class AdsExclusionSpecificationEntry(BaseExclusionSpecificationEntry):
  """Stores Google Ads specific rules."""

  def __init__(self, expression):
    super().__init__(
      expression, exclusion_type=enums.ExclusionTypeEnum.GOOGLE_ADS_INFO
    )


class ContentExclusionSpecificationEntry(BaseExclusionSpecificationEntry):
  """Stores Website specific rules."""

  def __init__(self, expression) -> None:
    super().__init__(
      expression=expression, exclusion_type=enums.ExclusionTypeEnum.WEBSITE_INFO
    )


class YouTubeChannelExclusionSpecificationEntry(
  BaseExclusionSpecificationEntry
):
  """Stores YouTube Channel specific rules."""

  def __init__(self, expression) -> None:
    super().__init__(
      expression=expression,
      exclusion_type=enums.ExclusionTypeEnum.YOUTUBE_CHANNEL_INFO,
    )


class YouTubeVideoExclusionSpecificationEntry(BaseExclusionSpecificationEntry):
  """Stores YouTube Video specific rules."""

  def __init__(self, expression) -> None:
    super().__init__(
      expression=expression,
      exclusion_type=enums.ExclusionTypeEnum.YOUTUBE_VIDEO_INFO,
    )


def create_exclusion_specification_entry(
  specification_type: str, condition: str
) -> BaseExclusionSpecificationEntry:
  """Builds concrete specification entry class based on type.

  Args:
    specification_type: Type of desired specification entry.
    condition: Expression to use for building the specification entry.

  Returns:
    Any subclass of instance of BaseExclusionSpecificationEntry.
  """
  if specification_type == 'GOOGLE_ADS_INFO':
    return AdsExclusionSpecificationEntry(condition)
  if specification_type == 'WEBSITE_INFO':
    return ContentExclusionSpecificationEntry(condition)
  if specification_type == 'YOUTUBE_CHANNEL_INFO':
    return YouTubeChannelExclusionSpecificationEntry(condition)
  if specification_type == 'YOUTUBE_VIDEO_INFO':
    return YouTubeVideoExclusionSpecificationEntry(condition)
  raise ValueError(f'Incorrect type of rule: {specification_type}')


class ExclusionSpecification:
  """Verifies whether entities matches set of specification entries.

  Attributes:
    specifications: All specification entries within the specification.
  """

  def __init__(
    self,
    specifications: Sequence[Sequence[BaseExclusionSpecificationEntry]]
    | None = None,
  ) -> None:
    """Initializes ExclusionSpecification.

    Args:
      specifications: All specification entries within the specification.
    """
    self.specifications = specifications

  def apply_specifications(
    self,
    placements: report.GaarfReport,
    include_reason: bool = True,
    include_matching_placement: bool = True,
  ) -> report.GaarfReport:
    """Gets placements that satisfy exclusion specifications entries.

    Args:
      placements:
        Report to be checked against specifications.
      include_reason:
        Whether to include exclusion reason to output report.
      include_matching_placement:
        Whether to include matching placement to output report.

    Returns:
      Report filtered to placements that matches all
      specification entries.
    """
    if not self.specifications:
      return placements
    to_be_excluded_placements = []
    extra_columns: list[str] = []
    if include_reason:
      extra_columns.append('reason')
    if include_matching_placement:
      extra_columns.append('matching_placement')
    for placement in placements:
      reason, matching_placement = self.satisfies(placement)
      extra_data: list = []
      if reason:
        if include_reason:
          reason_str = ','.join(list(itertools.chain(*reason)))
          extra_data = [reason_str]
          if include_matching_placement:
            extra_data.append(matching_placement)
        to_be_excluded_placements.append(placement.data + extra_data)
    return report.GaarfReport(
      results=to_be_excluded_placements,
      column_names=placements.column_names + extra_columns,
    )

  def satisfies(self, placement: report.GaarfRow) -> tuple[list[str], dict]:
    """Verifies whether a single entity satisfies the specifications.

    Args:
      placement: GaarfRow object with placement data.

    Returns:
      Tuple with list of rules that placement satisfies and
      placement itself.
    """
    rules_satisfied: list[str] = []
    placement_satisfied: dict = {}
    for spec_entry in self.specifications:
      spec_satisfied: list[str] = []
      for spec in spec_entry:
        if spec.exclusion_type != enums.ExclusionTypeEnum.GOOGLE_ADS_INFO:
          formatted_exclusion_type = spec.exclusion_type.name.replace(
            '_INFO', ''
          )
          if formatted_exclusion_type != placement.placement_type:
            continue
        is_satisfied, placement_satisfied = spec.is_satisfied_by(placement)
        if is_satisfied:
          spec_satisfied.append(str(spec))
          continue
      if len(spec_satisfied) == len(spec_entry):
        rules_satisfied.append(spec_satisfied)
    return rules_satisfied, placement_satisfied

  @property
  def ads_specs_entries(self) -> ExclusionSpecification:
    """Specification filtered to Ads specific specification entries."""
    return self._get_specification_subset(
      include_exclusion_types={
        enums.ExclusionTypeEnum.GOOGLE_ADS_INFO,
      }
    )

  @property
  def non_ads_specs_entries(self) -> ExclusionSpecification:
    """Specification filtered to non-Ads specific specification entries."""
    return self._get_specification_subset(
      exclude_exclusion_types={
        enums.ExclusionTypeEnum.GOOGLE_ADS_INFO,
      }
    )

  @property
  def parsable_spec_entries(self) -> ExclusionSpecification:
    """Specification filtered to parsable specification entries."""
    return self._get_specification_subset(
      include_exclusion_types=_PARSEABLE_PLACEMENT_TYPES_ENUMS
    )

  def _get_specification_subset(
    self,
    include_exclusion_types: set[enums.ExclusionTypeEnum] | None = None,
    exclude_exclusion_types: set[enums.ExclusionTypeEnum] | None = None,
  ) -> ExclusionSpecification:
    """Builds new specification from a subset.

    Args:
      include_exclusion_types:
        Set of exclusion to include into new specifications.
      exclude_exclusion_types:
        Set of exclusion to include from new specifications.

    Returns:
      New ExclusionSpecification.
    """
    if not self.specifications:
      return ExclusionSpecification()
    specifications: list[list[BaseExclusionSpecificationEntry]] = []
    if include_exclusion_types:
      allowed_exclusion_types = include_exclusion_types
    else:
      exclude_exclusion_types = exclude_exclusion_types or set()
      allowed_exclusion_types = set(enums.ExclusionTypeEnum).difference(
        exclude_exclusion_types
      )
    for specification in self.specifications:
      matching_specification = []
      for specification_entry in specification:
        if specification_entry.exclusion_type in allowed_exclusion_types:
          matching_specification.append(specification_entry)
      if matching_specification:
        specifications.append(matching_specification)
    return ExclusionSpecification(specifications=specifications)

  def define_runtime_options(self) -> RuntimeOptions:
    """Generates runtime options based on provided specifications."""
    if not self.specifications:
      return RuntimeOptions()
    is_conversion_query = False
    conversion_name = ''
    conversion_rules: list[BaseExclusionSpecificationEntry] = []
    for specification in self.specifications:
      for rule in specification:
        if rule.name == 'conversion_name':
          is_conversion_query = True
          conversion_name = rule.value
          conversion_rules.append(rule)
    return RuntimeOptions(
      is_conversion_query=is_conversion_query,
      conversion_name=conversion_name,
      conversion_rules=conversion_rules,
    )

  @classmethod
  def from_rules(
    cls, parsed_rules: list[list[rules_parser.Rule]] | None
  ) -> ExclusionSpecification:
    """Convert Exclusion rules into specifications."""
    if not parsed_rules:
      return ExclusionSpecification()
    specifications = []
    for rules in parsed_rules:
      specification_entry = []
      for rule in rules:
        specification_entry.append(
          create_exclusion_specification_entry(
            rule.exclusion_type, rule.exclusion_rule
          )
        )
      specifications.append(specification_entry)
    return ExclusionSpecification(specifications)

  @classmethod
  def from_expression(cls, rule_expression: str) -> ExclusionSpecification:
    """Convert raw string rules into specifications."""
    return ExclusionSpecification.from_rules(
      rules_parser.generate_rules(rule_expression)
    )

  def __eq__(self, other) -> bool:
    return self.specifications == other.specifications

  def __bool__(self) -> bool:
    return bool(self.specifications)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for performing parsing of external entities.

External entity represents a YouTube Channel or Video, Webpage,
Mobile Application or anything that requires connection to external
(non Google Ads) system in order to get information from it.
"""

from __future__ import annotations

import dataclasses
import itertools
import logging

from gaarf import report

from googleads_housekeeper.adapters import repository
from googleads_housekeeper.domain import external_parsers
from googleads_housekeeper.domain.core import exclusion_specification
from googleads_housekeeper.services import unit_of_work

PARSER_MAPPING = {
  'youtube_channel_info': external_parsers.youtube_data_parser.ChannelInfoParser,
  'youtube_video_info': external_parsers.youtube_data_parser.VideoInfoParser,
  'website_info': external_parsers.website_parser.WebSiteParser,
}


@dataclasses.dataclass
class ParseOptions:
  """Specifies options regarding parsing behaviour.

  Attributes:
      save_to_db: Whether to save parsing results to DB.
      batch_size: Maximum number of entities in response.
  """

  save_to_db: bool = True
  batch_size: int = 50


class ExternalEntitiesParser:
  """Performance parsing of external entities (YouTube, Websites, etc.).

  Attributes:
      uow: Unit of work to handle transaction.
  """

  def __init__(self, uow: unit_of_work.AbstractUnitOfWork) -> None:
    """Initializes an instance of ExternalEntitiesParser.

    Args:
        uow: Unit of work to handle transaction.
    """
    self.uow = uow

  def parse_specification_chain(
    self,
    entities: report.GaarfReport,
    specification: exclusion_specification.ExclusionSpecification,
    parse_options: ParseOptions = ParseOptions(),
  ) -> None:
    """Performance parsing of entities via all parsers in specification.

    Args:
        entities:
            Report containing entities that should be parsed.
        specifications:
            All possible non Ads specifications. Entities within a report
            can be parsed via different parsers (YouTube, Website, etc.)
        parse_options:
            Options for performing parsing.
    """
    with self.uow as uow:
      for specification_entries in specification.specifications:
        for specification_entry in specification_entries:
          self._parse_via_external_parser(
            entities=entities,
            specification_entry=specification_entry,
            uow=uow,
            parse_options=parse_options,
          )

  def _parse_via_external_parser(
    self,
    entities: report.GaarfReport,
    specification_entry: exclusion_specification.BaseExclusionSpecificationEntry,
    uow: unit_of_work.AbstractUnitOfWork,
    parse_options: ParseOptions,
  ) -> None:
    """Performance parsing of entities via a single parser in specification.

    Args:
        entities:
            Report containing entities that should be parsed.
        specification:
            Single non Ads specifications. Entities will be parsed via
            a single parsers (YouTube, Website, etc.
        uow:
            Unit of work to handle transaction.
        parse_options:
            Options for performing parsing.
    """
    repo = getattr(uow, specification_entry.repository_name)
    parser = PARSER_MAPPING.get(specification_entry.repository_name)
    if not_parsed_entities := self._get_not_parsed_entities(
      entities, specification_entry, repo
    ):
      i = 0
      while batch := list(
        itertools.islice(not_parsed_entities, i, i + parse_options.batch_size)
      ):
        parsed_entities_info = parser().parse(batch)
        i += parse_options.batch_size
        if parse_options.save_to_db:
          for parsed_placement in parsed_entities_info:
            logging.debug('saving placement: %s to db', parsed_placement)
            repo.add(parsed_placement)
          uow.commit()

  def _get_not_parsed_entities(
    self,
    entities: report.GaarfReport,
    specification_entry: exclusion_specification.BaseExclusionSpecificationEntry,
    repo: repository.AbstractRepository,
  ) -> list[str]:
    """Get all entities not parsed and saved in DB.

    Args:
        entities:
            Report containing entities that should be parsed.
        specification_entry:
            Single non Ads specifications. Entities will be parsed via
            a single parsers (YouTube, Website, etc.).
        repo:
            Repository to get data from.

    Returns:
        Non-parsed placements.
    """
    not_parsed_entities: list[str] = []
    for entity_info in entities:
      if (
        entity_info.placement_type
        == specification_entry.corresponding_placement_type.name
      ):
        if not repo.get_by_conditions(
          {
            'placement': entity_info.placement,
          }
        ):
          not_parsed_entities.append(str(entity_info.placement))
    return not_parsed_entities



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining publishers.

Publisher connects Ads Housekeeper with external systems.
"""

from __future__ import annotations

import dataclasses
import json
import logging

from typing_extensions import override

from googleads_housekeeper.domain import events


class BasePublisher:
  """Interface to inherit all publishers from."""

  def publish(self, topic: str, event: events.Event) -> None:
    """Publishes event to a topic."""


class RedisPublisher(BasePublisher):
  """Publishes messages to Redis.

  Attributes:
    client: Instantiated Redis client.
    topic_prefix: Prefix to identify topics related to the application.
  """

  def __init__(
    self, client: 'redis.Redis', topic_prefix: str | None = None
  ) -> None:
    """Initializes RedisPublisher.

    Args:
      client: Instantiated Redis client.
      topic_prefix: Prefix to identify topics related to the application.
    """
    self.client = client
    self.topic_prefix = topic_prefix

  @override
  def publish(self, topic: str, event: events.Event) -> None:
    if self.topic_prefix:
      topic = f'{self.topic_prefix}_{topic}'
    self.client.publish(
      topic, json.dumps(dataclasses.asdict(event), default=str)
    )


class GoogleCloudPubSubPublisher(BasePublisher):
  """Publishes messages to Google Cloud PubSub.

  Attributes:
    client: Instantiated Pubsub client.
    topic_prefix: Prefix to identify topics related to the application.
  """

  def __init__(
    self,
    client: 'google.cloud.pubsub_v1.PublisherClient',
    project_id: str,
    topic_prefix: str | None = None,
  ) -> None:
    """Initializes GoogleCloudPubSubPublisher.

    Args:
      client: Instantiated Pubsub client.
      project_id: Google Cloud project identifier.
      topic_prefix: Prefix to identify topics related to the application.
    """
    self.client = client
    self.project_id = project_id
    self.topic_prefix = topic_prefix

  @override
  def publish(self, topic: str, event: events.Event) -> None:
    if self.topic_prefix:
      topic = f'{self.topic_prefix}_{topic}'
    topic_name = f'projects/{self.project_id}/topics/{topic}'
    data = str(json.dumps(dataclasses.asdict(event), default=str)).encode(
      'utf-8'
    )
    future = self.client.publish(topic_name, data=data)
    future.result()


class LogPublisher(BasePublisher):
  """Sends message to a log.

  Attributes:
    topic_prefix: Prefix to identify topics related to the application.
  """

  def __init__(self, topic_prefix: str | None = None) -> None:
    """Initializes LogPublisher.

    Args:
      topic_prefix: Prefix to identify topics related to the application.
    """
    self.topic_prefix = topic_prefix

  @override
  def publish(self, topic: str, event: events.Event) -> None:
    if self.topic_prefix:
      topic = f'{self.topic_prefix}_{topic}'
    logging.info(
      "Published to topic '%s': %s", topic, dataclasses.asdict(event)
    )



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from collections import abc
from dataclasses import dataclass

from googleads_housekeeper.domain.core import execution, task


class Command: ...


@dataclass
class RunTask(Command):
  id: int
  type: execution.ExecutionTypeEnum
  exclusion_rule: str = ''
  save_to_db: bool = True


@dataclass
class SaveTask(Command):
  exclusion_rule: str
  customer_ids: str | list[str]
  from_days_ago: int = 0
  date_range: int = 7
  exclusion_level: str = 'AD_GROUP'
  output: str = task.TaskOutput.EXCLUDE_AND_NOTIFY.name
  name: str | None = None
  schedule: str | None = None
  placement_types: str | None = None
  task_id: str | None = None

  def __post_init__(self):
    self.customer_ids = (
      ','.join(self.customer_ids)
      if isinstance(self.customer_ids, abc.MutableSequence)
      else self.customer_ids
    )


@dataclass
class DeleteTask(Command):
  task_id: int


@dataclass
class RunManualExclusion(Command):
  customer_ids: str
  placements: list[list]
  header: list[str]
  exclusion_level: str


def ensure_tuple(value):
  if isinstance(value, str):
    return tuple(value.split(','))
  return value


@dataclass
class PreviewPlacements(Command):
  exclusion_rule: str
  placement_types: tuple[str, ...] | None
  customer_ids: str | list[str]
  from_days_ago: int
  date_range: int
  exclusion_level: str = 'AD_GROUP'
  exclude_and_notify: str = 'EXCLUDE_AND_NOTIFY'
  save_to_db: bool = True
  preview_pagination_index_one_based: int = 1

  def __post_init__(self):
    self.placement_types = ensure_tuple(self.placement_types)
    self.from_days_ago = int(self.from_days_ago)
    self.date_range = int(self.date_range)


@dataclass
class AddToAllowlisting(Command):
  type: str
  name: str
  account_id: str


@dataclass
class RemoveFromAllowlisting(Command):
  type: str
  name: str
  account_id: str


@dataclass
class SaveConfig(Command):
  id: str
  mcc_id: str
  email_address: str
  always_fetch_youtube_preview_mode: bool = True
  save_to_db: bool = False


@dataclass
class GetCustomerIds(Command):
  mcc_id: str


@dataclass
class GetMccIds(Command):
  root_mcc_id: int


@dataclass
class SaveWebsiteInfo(Command):
  website_info: dict


@dataclass
class SaveChannelInfo(Command):
  channel_info: dict


@dataclass
class SaveVideoInfo(Command):
  video_info: dict


@dataclass
class MigrateFromOldTasks(Command):
  pass



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Application handlers."""

from __future__ import annotations

import json
import logging
from copy import copy
from dataclasses import asdict
from datetime import datetime
from typing import Any

from gaarf.api_clients import GoogleAdsApiClient
from gaarf.query_executor import AdsReportFetcher
from gaarf.report import GaarfReport

from googleads_housekeeper.adapters.legacy_adapter import TaskAdapter
from googleads_housekeeper.adapters.notifications import MessagePayload
from googleads_housekeeper.adapters.publisher import BasePublisher
from googleads_housekeeper.domain import commands, events
from googleads_housekeeper.domain.core import execution, settings, task
from googleads_housekeeper.domain.placement_handler import (
  entities as placement_entities,
)
from googleads_housekeeper.domain.placement_handler import (
  placement_excluder,
  placement_fetcher,
  placement_service,
)
from googleads_housekeeper.services import enums, unit_of_work


def get_mcc_accounts(
  cmd: commands.GetMccIds,
  uow: unit_of_work.AbstractUnitOfWork,
  ads_api_client: GoogleAdsApiClient,
) -> list[dict[str, str]]:
  """Fetches and updates mcc accounts in the repository.

  Args:
      cmd: command for getting MccIds.
      uow: unit of work to handle the transaction.
      ads_api_client: client to perform report fetching.

  Returns:
      All mcc accounts available under given root_mcc_id.
  """
  mccs = get_accessible_mccs(ads_api_client, cmd.root_mcc_id)
  result: list[dict[str, str]] = []
  with uow:
    saved_mcc_ids = {int(account.id) for account in uow.mcc_ids.list()}
    fetched_mcc_ids = set(
      mccs['account_id'].to_list(flatten=True, distinct=True)
    )
    new_accounts = fetched_mcc_ids.difference(saved_mcc_ids)
    to_be_removed_accounts = saved_mcc_ids.difference(fetched_mcc_ids)
    for row in mccs:
      row_dict = {'id': row.account_id, 'account_name': row.account_name}
      result.append(row_dict)
      if row.account_id in new_accounts:
        uow.mcc_ids.add(settings.MccIds(**row_dict))
    for account_id in to_be_removed_accounts:
      uow.mcc_ids.delete(account_id)
    uow.commit()
  return result


def get_accessible_mccs(
  ads_api_client: GoogleAdsApiClient, root_mcc_id: int
) -> GaarfReport:
  """Fetches all available mccs under a root_mcc_id.

  Args:
      ads_api_client: client to perform report fetching.
      root_mcc_id: root mmc id for which leaf mccs should be fetched.

  Returns:
  Report with mcc names and ids.
  """
  query = """
        SELECT
            customer_client.descriptive_name AS account_name,
            customer_client.id AS account_id
        FROM customer_client
        WHERE customer_client.manager = TRUE
        AND customer_client.status = "ENABLED"
        """
  report_fetcher = AdsReportFetcher(api_client=ads_api_client)
  return report_fetcher.fetch(query, root_mcc_id)


def get_customer_ids(
  cmd: commands.GetCustomerIds,
  uow: unit_of_work.AbstractUnitOfWork,
  ads_api_client: GoogleAdsApiClient,
) -> list[dict[str, int]]:
  """Fetches accounts under a given mcc_id and update them in the repository.

  Args:
      cmd: command for getting customer ids under a given mcc_id.
      uow: unit of work to handle the transaction.
      ads_api_client: client to perform report fetching.

  Returns:
      All accounts available under given mcc_id.
  """
  customer_ids_query = """
    SELECT
        customer_client.descriptive_name AS account_name,
        customer_client.id AS account_id
    FROM customer_client
    WHERE customer_client.manager = FALSE
    AND customer_client.status = "ENABLED"
    """
  report_fetcher = AdsReportFetcher(api_client=ads_api_client)
  customer_ids = report_fetcher.fetch(
    customer_ids_query, customer_ids=cmd.mcc_id
  )
  saved_customer_ids = set(
    [
      int(account.id)
      for account in uow.customer_ids.get_by_conditions({'mcc_id': cmd.mcc_id})
    ]
  )
  fetched_customer_ids = set(
    customer_ids['account_id'].to_list(flatten=True, distinct=True)
  )
  new_accounts = fetched_customer_ids.difference(saved_customer_ids)
  to_be_removed_accounts = saved_customer_ids.difference(fetched_customer_ids)
  result: list[dict[str, int]] = []
  with uow:
    for row in customer_ids:
      result.append({'account_name': row.account_name, 'id': row.account_id})
      if row.account_id in new_accounts:
        uow.customer_ids.add(
          settings.CustomerIds(
            mcc_id=cmd.mcc_id,
            account_name=row.account_name,
            id=str(row.account_id),
          )
        )
    for account_id in to_be_removed_accounts:
      uow.customer_ids.delete(account_id)
    uow.commit()
  return result


def run_manual_exclusion_task(
  cmd: commands.RunManualExclusion,
  uow: unit_of_work.AbstractUnitOfWork,
  ads_api_client: GoogleAdsApiClient,
) -> dict[str, int]:
  with uow:
    report_fetcher = AdsReportFetcher(ads_api_client)
    excluder = placement_excluder.PlacementExcluder(
      ads_api_client, cmd.exclusion_level
    )
    fetcher = placement_fetcher.PlacementFetcher(report_fetcher)
    placement_exclusion_lists = fetcher.get_placement_exclusion_lists(
      cmd.customer_ids
    )
    exclusion_result = excluder.exclude_placements(
      to_be_excluded_placements=GaarfReport(
        results=cmd.placements, column_names=cmd.header
      ),
      placement_exclusion_lists=placement_exclusion_lists,
    )
    n_excluded_placements = len(exclusion_result.excluded_placements or [])
    n_associated_with_list_placements = len(
      exclusion_result.associated_with_list_placements or []
    )
    n_excludable_from_account_only = len(
      exclusion_result.excludable_from_account_only or []
    )
    return {
      'excluded_placements': n_excluded_placements,
      'associated_with_list_placements': n_associated_with_list_placements,
      'excludable_from_account_only': n_excludable_from_account_only,
    }


def task_created(event: events.TaskCreated, publisher: BasePublisher) -> None:
  publisher.publish('task_created', event)


def task_with_schedule_created(
  event: events.TaskWithScheduleCreated, publisher: BasePublisher
) -> None:
  publisher.publish('task_created', event)


def task_updated(event: events.TaskUpdated, publisher: BasePublisher) -> None:
  publisher.publish('task_updated', event)


def task_schedule_updated(
  event: events.TaskScheduleUpdated, publisher: BasePublisher
) -> None:
  publisher.publish('task_updated', event)


def task_deleted(event: events.TaskDeleted, publisher: BasePublisher) -> None:
  publisher.publish('task_deleted', event)


def task_schedule_deleted(
  event: events.TaskScheduleDeleted, publisher: BasePublisher
) -> None:
  publisher.publish('task_deleted', event)


def run_task(
  cmd: commands.RunTask,
  uow: unit_of_work.AbstractUnitOfWork,
  ads_api_client: GoogleAdsApiClient,
  publisher: BasePublisher | None = None,
  save_to_db: bool = True,
) -> tuple[dict, MessagePayload | None]:
  with copy(uow) as uow:
    settings = uow.settings.list()
    task_obj = uow.tasks.get(task_id=cmd.id)
    start_time = datetime.now()
    exclusion_result = placement_service.exclude_placements(
      task_obj=task_obj,
      uow=uow,
      client=ads_api_client,
      runtime_options=placement_service.RuntimeOptions(
        always_fetch_youtube_preview_mode=False
      ),
    )
    end_time = datetime.now()
    n_excluded_placements = len(exclusion_result.excluded_placements or [])
    n_associated_with_list_placements = len(
      exclusion_result.associated_with_list_placements or []
    )
    execution_obj = execution.Execution(
      task=cmd.id,
      start_time=start_time,
      end_time=end_time,
      placements_excluded=n_excluded_placements,
      type=cmd.type,
    )
    uow.executions.add(execution_obj)
    if save_to_db and exclusion_result.excluded_placements:
      for placement in exclusion_result.excluded_placements:
        if hasattr(placement, 'reason'):
          exclusion_reason = placement.reason
        else:
          exclusion_reason = ''
        uow.execution_details.add(
          execution.ExecutionDetails(
            execution_id=execution_obj.id,
            placement=placement.name,
            placement_type=placement.placement_type,
            reason=exclusion_reason,
          )
        )
    uow.commit()
    if task_obj.output in (
      task.TaskOutput.NOTIFY,
      task.TaskOutput.EXCLUDE_AND_NOTIFY,
    ):
      if excluded_placements := exclusion_result.excluded_placements:
        message_payload = MessagePayload(
          task_id=task_obj.id,
          task_name=task_obj.name,
          task_formula=task_obj.exclusion_rule,
          task_output=task_obj.output,
          placements_excluded_sample=excluded_placements[0:10],
          total_placement_excluded=len(excluded_placements),
          recipient=settings[0].email_address,
        )
      else:
        message_payload = MessagePayload(
          task_id=task_obj.id,
          task_name=task_obj.name,
          task_formula=task_obj.exclusion_rule,
          task_output=task_obj.output,
          placements_excluded_sample=None,
          total_placement_excluded=0,
          recipient=settings[0].email_address,
        )
    else:
      message_payload = None
    return {
      'excluded_placements': n_excluded_placements,
      'associated_with_list_placements': n_associated_with_list_placements,
    }, message_payload


def preview_placements(
  cmd: commands.PreviewPlacements,
  uow: unit_of_work.AbstractUnitOfWork,
  ads_api_client: GoogleAdsApiClient,
  always_fetch_youtube_preview_mode: bool = True,
  save_to_db: bool = True,
) -> dict[str, Any]:
  report_fetcher = AdsReportFetcher(api_client=ads_api_client)
  task_obj = task.Task(
    name='',
    exclusion_rule=cmd.exclusion_rule,
    customer_ids=cmd.customer_ids,
    date_range=cmd.date_range,
    from_days_ago=cmd.from_days_ago,
    exclusion_level=enums.ExclusionLevelEnum[cmd.exclusion_level],
    placement_types=cmd.placement_types,
  )
  to_be_excluded_placements = placement_service.find_placements_for_exclusion(
    task_obj=task_obj,
    uow=uow,
    report_fetcher=report_fetcher,
    preview_pagination_index_zero_based=cmd.preview_pagination_index_one_based
    - 1,
    runtime_options=placement_service.RuntimeOptions(
      always_fetch_youtube_preview_mode=always_fetch_youtube_preview_mode,
      save_to_db=save_to_db,
    ),
  )
  if not to_be_excluded_placements:
    data = {}
  else:
    data = json.loads(
      to_be_excluded_placements.to_pandas().to_json(orient='index')
    )
  return {
    'data': data,
    'dates': {'date_from': task_obj.start_date, 'date_to': task_obj.end_date},
  }


def save_task(
  cmd: commands.SaveTask,
  uow: unit_of_work.AbstractUnitOfWork,
) -> str:
  with uow:
    task_id = None
    task_obj = None
    if hasattr(cmd, 'task_id') and cmd.task_id:
      task_id = cmd.task_id
      task_obj = uow.tasks.get(task_id=task_id)
      update_dict = asdict(cmd)
      update_dict.pop('task_id')
      uow.tasks.update(task_obj.id, update_dict)
      uow.commit()
      uow.published_events.append(events.TaskUpdated(cmd.task_id))
      if cmd.schedule and cmd.schedule != task_obj.schedule:
        uow.published_events.append(
          events.TaskScheduleUpdated(cmd.task_id, cmd.schedule)
        )
      else:
        uow.published_events.append(events.TaskScheduleDeleted(cmd.task_id))
    else:
      task_dict = asdict(cmd)
      task_dict.pop('task_id')
      task_obj = task.Task(**task_dict)
      uow.tasks.add(task_obj)
      uow.commit()
      task_id = task_obj.id
      uow.published_events.append(events.TaskCreated(task_id))
      if cmd.schedule:
        uow.published_events.append(
          events.TaskWithScheduleCreated(task_id, cmd.name, cmd.schedule)
        )
    task_id = task_obj.id
    return str(task_id)


def delete_task(
  cmd: commands.DeleteTask,
  uow: unit_of_work.AbstractUnitOfWork,
):
  with copy(uow) as uow:
    task_obj = uow.tasks.get(task_id=cmd.task_id)
    if task_obj:
      uow.tasks.update(cmd.task_id, {'status': 'INACTIVE'})
      uow.commit()
      uow.published_events.append(events.TaskDeleted(cmd.task_id))
      uow.published_events.append(events.TaskScheduleDeleted(cmd.task_id))
    else:
      logging.warning('No task with id %d found!', cmd.id)


def save_config(
  cmd: commands.SaveConfig,
  uow: unit_of_work.AbstractUnitOfWork,
):
  with uow:
    if hasattr(cmd, 'id') and cmd.id:
      config_id = cmd.id
      config = uow.settings.get(config_id)
      update_dict = asdict(cmd)
      update_dict.pop('id')
      uow.settings.update(config.id, update_dict)
    else:
      config_dict = asdict(cmd)
      config_dict.pop('id')
      config = settings.Config(**config_dict)
      uow.settings.add(config)
    uow.commit()


def add_to_allowlisting(
  cmd: commands.AddToAllowlisting, uow: unit_of_work.AbstractUnitOfWork
) -> None:
  with uow:
    if not uow.allowlisting.get_by_conditions(asdict(cmd)):
      placement = placement_entities.AllowlistedPlacement(**asdict(cmd))
      uow.allowlisting.add(placement)
      uow.commit()


def remove_from_allowlisting(
  cmd: commands.RemoveFromAllowlisting, uow: unit_of_work.AbstractUnitOfWork
) -> None:
  with uow:
    if allowlisted_placement := uow.allowlisting.get_by_conditions(asdict(cmd)):
      uow.allowlisting.delete(allowlisted_placement[0].id)
      uow.commit()


def save_channel_info(
  cmd: commands.SaveChannelInfo, uow: unit_of_work.AbstractUnitOfWork
) -> None:
  with uow:
    uow.youtube_channel_info.add(cmd.channel_info)
    uow.commit()


def save_video_info(
  cmd: commands.SaveChannelInfo, uow: unit_of_work.AbstractUnitOfWork
) -> None:
  with uow:
    uow.youtube_video_info.add(cmd.video_info)
    uow.commit()


def save_website_info(
  cmd: commands.SaveChannelInfo, uow: unit_of_work.AbstractUnitOfWork
) -> None:
  with uow:
    uow.website_info.add(cmd.website_info)
    uow.commit()


def migrate_from_old_tasks(
  cmd: commands.MigrateFromOldTasks, uow: unit_of_work.AbstractUnitOfWork
) -> None:
  with uow:
    for item in uow.old_tasks.list():
      updated_item_dict = TaskAdapter(item.to_dict()).from_task_v1().to_dict()
      uow.tasks.update(item.id, updated_item_dict)


EVENT_HANDLERS = {
  events.TaskWithScheduleCreated: [task_with_schedule_created],
  events.TaskScheduleUpdated: [task_schedule_updated],
  events.TaskScheduleDeleted: [task_schedule_deleted],
  events.TaskCreated: [task_created],
  events.TaskUpdated: [task_updated],
  events.TaskDeleted: [task_deleted],
}

COMMAND_HANDLERS = {
  commands.RunTask: run_task,
  commands.SaveTask: save_task,
  commands.DeleteTask: delete_task,
  commands.RunManualExclusion: run_manual_exclusion_task,
  commands.PreviewPlacements: preview_placements,
  commands.SaveConfig: save_config,
  commands.GetCustomerIds: get_customer_ids,
  commands.GetMccIds: get_mcc_accounts,
  commands.AddToAllowlisting: add_to_allowlisting,
  commands.RemoveFromAllowlisting: remove_from_allowlisting,
  commands.SaveChannelInfo: save_channel_info,
  commands.SaveVideoInfo: save_video_info,
  commands.SaveWebsiteInfo: save_website_info,
  commands.MigrateFromOldTasks: migrate_from_old_tasks,
}



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import inspect
import os

from gaarf.api_clients import BaseClient, GoogleAdsApiClient

if os.getenv('ADS_HOUSEKEEPER_DEPLOYMENT_TYPE') == 'Google Cloud':
  try:
    pass
  except ImportError as e:
    raise ImportError(
      'Please install googleads-housekeeper with Google Cloud support - `pip install googleads-housekeeper[gcp]`'
    ) from e

from googleads_housekeeper.adapters import notifications, orm, publisher
from googleads_housekeeper.services import handlers, messagebus, unit_of_work


class Bootstrapper:
  def __init__(
    self, type: str = 'Local', topic_prefix: str | None = None
  ) -> None:
    self.notification_service = notifications.ConsoleNotifications()
    if type == 'Local':
      self.bootstrapper = LocalBootstrapper(topic_prefix=topic_prefix)
    if type == 'Google Cloud':
      self.bootstrapper = GoogleCloudBootstrapper(topic_prefix=topic_prefix)
    if type == 'Dev':
      self.bootstrapper = DevBootstrapper(topic_prefix=topic_prefix)

  def bootstrap_app(
    self,
    ads_api_client=None,
    start_orm=False,
    uow=None,
    notification_service=None,
    publish_service=None,
  ):
    return self.bootstrapper._bootstrap(
      ads_api_client=ads_api_client,
      start_orm=start_orm,
      uow=uow,
      notification_service=notification_service,
      publish_service=publish_service,
    )

  def _bootstrap(
    self, ads_api_client, start_orm, uow, notification_service, publish_service
  ):
    return bootstrap(
      ads_api_client=ads_api_client,
      start_orm=start_orm or self.start_orm,
      uow=uow or self.uow,
      notification_service=notification_service or self.notification_service,
      publish_service=publish_service or self.publish_service,
    )


class GoogleCloudBootstrapper(Bootstrapper):
  def __init__(
    self,
    project_id: str = os.getenv('GOOGLE_CLOUD_PROJECT'),
    appengine_region: str = os.getenv('APPENGINE_REGION', 'europe-west1'),
    start_orm: bool = False,
    topic_prefix: str | None = None,
  ) -> None:
    from google.cloud import firestore, pubsub_v1

    self.start_orm = False
    self.uow = unit_of_work.FirestoreUnitOfWork(
      firestore.Client(project=project_id)
    )
    self.notification_service = (
      notifications.GoogleCloudAppEngineEmailNotifications(
        project_id=project_id, appengine_region=appengine_region
      )
    )
    self.publish_service = publisher.GoogleCloudPubSubPublisher(
      client=pubsub_v1.PublisherClient(),
      project_id=project_id,
      topic_prefix=topic_prefix,
    )


class LocalBootstrapper(Bootstrapper):
  def __init__(self, topic_prefix: str | None = None) -> None:
    self.start_orm = True
    self.uow = unit_of_work.SqlAlchemyUnitOfWork(get_db_uri())
    self.publish_service = publisher.LogPublisher(topic_prefix)
    self.notification_service = notifications.ConsoleNotifications()


class DevBootstrapper(Bootstrapper):
  def __init__(self, topic_prefix: str | None = None) -> None:
    self.start_orm = True
    self.uow = unit_of_work.SqlAlchemyUnitOfWork(
      f'sqlite:///{os.getcwd()}/ads_housekeeper_tasks.db'
    )
    self.publish_service = publisher.LogPublisher(topic_prefix)
    self.notification_service = notifications.ConsoleNotifications()


def get_db_uri(
  host: str = os.environ.get('DB_HOST', 'localhost'),
  port: int = int(os.environ.get('DB_PORT', 54321)),
  user: str = os.environ.get('DB_USER', 'ads_housekeeper'),
  password: str = os.environ.get('DB_PASSWORD', 'ads_housekeeper'),
  db_name: str = os.environ.get('DB_NAME', 'ads_housekeeper'),
  db_engine: str = os.environ.get('DB_ENGINE', 'postgresql'),
) -> str:
  if dev_db := os.environ.get('DEV_DB_HOST'):
    return f'sqlite:///{dev_db}'
  return f'{db_engine}://{user}:{password}@{host}:{port}/{db_name}'


def get_default_google_ads_api_client():
  return GoogleAdsApiClient(
    path_to_config=get_path_to_ads_config(), version=get_ads_api_version()
  )


def get_ads_api_version() -> str:
  return os.environ.get('GOOGLE_ADS_API_VERSION', 'v16')


def get_path_to_ads_config() -> str:
  return os.environ.get('GOOGLE_ADS_PATH_TO_CONFIG', '~/google-ads.yaml')


def get_notification_service() -> notifications.BaseNotifications:
  notification_service_factory = notifications.NotificationFactory()
  notification_type = os.environ.get(
    'ADS_HOUSEKEEPER_NOTIFICATION_TYPE', 'console'
  )
  notification_service = (
    notification_service_factory.create_notification_service(
      notification_type=notification_type
    )
  )
  return notification_service


def bootstrap(
  ads_api_client: BaseClient | None = None,
  start_orm: bool = True,
  uow: unit_of_work.AbstractUnitOfWork = unit_of_work.SqlAlchemyUnitOfWork(
    get_db_uri()
  ),
  notification_service: notifications.BaseNotifications = get_notification_service(),
  publish_service: publisher.BasePublisher = publisher.LogPublisher(),
) -> messagebus.MessageBus:
  if start_orm:
    orm.start_mappers(engine=uow.engine)

  dependencies = {
    'uow': uow,
    'ads_api_client': ads_api_client or get_default_google_ads_api_client(),
    'notification_service': notification_service,
    'publisher': publish_service,
  }
  injected_event_handlers = {
    event_type: [
      inject_dependencies(handler, dependencies) for handler in handlers
    ]
    for event_type, handlers in handlers.EVENT_HANDLERS.items()
  }
  injected_command_handlers = {
    command_type: inject_dependencies(handler, dependencies)
    for command_type, handler in handlers.COMMAND_HANDLERS.items()
  }
  return messagebus.MessageBus(
    uow=uow,
    event_handlers=injected_event_handlers,
    command_handlers=injected_command_handlers,
    dependencies=dependencies,
  )


def inject_dependencies(handler, dependencies):
  params = inspect.signature(handler).parameters
  deps = {
    name: dependency
    for name, dependency in dependencies.items()
    if name in params
  }
  return lambda message: handler(message, **deps)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import abc
from dataclasses import asdict
from typing import Any

from google.cloud.firestore_v1.base_query import FieldFilter

from googleads_housekeeper.domain.core import task


class AbstractRepository(abc.ABC):
  def __init__(self):
    self.seen = set()

  def add(self, task: task.Task):
    self._add(task)

  def delete(self, task_id: str):
    self._delete(task_id)

  def get(self, task_id) -> task.Task:
    return self._get(task_id)

  def get_by_conditions(self, conditions: dict[str, Any]) -> task.Task:
    return self._get_by_conditions(conditions)

  def get_by_condition(
    self, condition_name: str, condition_value: str
  ) -> task.Task:
    return self._get_by_conditions({condition_name: condition_value})

  def list(self) -> list[task.Task]:
    return self._list()

  def update(self, task_id: str, update_dict: dict[str, str]) -> task.Task:
    return self._update(task_id, update_dict)

  @abc.abstractmethod
  def _add(self, task: task.Task): ...

  @abc.abstractmethod
  def _get(self, task_id) -> task.Task: ...

  @abc.abstractmethod
  def _get_by_conditions(self, conditions: dict[str, Any]) -> task.Task: ...

  @abc.abstractmethod
  def _list(self) -> list[task.Task]: ...

  @abc.abstractmethod
  def _update(self, task_id, update_dict): ...

  @abc.abstractmethod
  def _delete(self, task_id): ...


class SqlAlchemyRepository(AbstractRepository):
  def __init__(self, session, entity=task.Task):
    super().__init__()
    self.session = session
    self.entity = entity

  def _add(self, task):
    self.session.add(task)

  def _get(self, task_id):
    return self.session.query(self.entity).filter_by(id=task_id).first()

  def _get_by_conditions(self, conditions: dict[str, Any]):
    query = self.session.query(self.entity)
    for condition_name, condition_value in conditions.items():
      query = query.filter(
        getattr(self.entity, condition_name) == condition_value
      )
    return query.all()

  def _list(self):
    return self.session.query(self.entity).all()

  def _update(self, task_id, update_dict):
    return (
      self.session.query(self.entity).filter_by(id=task_id).update(update_dict)
    )

  def _delete(self, task_id):
    return self.session.query(self.entity).filter_by(id=task_id).delete()


class FirestoreRepository(AbstractRepository):
  def __init__(self, client, entity=task.Task):
    super().__init__()
    self.client = client
    self.entity = entity
    self.collection_name = entity.__name__

  def _add(self, task):
    if hasattr(task, 'id'):
      element_id = task.id
    else:
      element_id = task._id
    element_dict = {}
    for key, value in asdict(task).items():
      if hasattr(value, 'name'):
        value = value.name
      element_dict[key] = value
    self.client.collection(self.collection_name).document(str(element_id)).set(
      element_dict
    )

  def _get(self, task_id: str):
    doc = self.client.collection(self.collection_name).document(task_id).get()
    if doc.exists:
      return self.entity(**doc.to_dict())
    return None

  def _get_by_conditions(self, conditions: dict[str, Any]):
    try:
      query = self.client.collection(self.collection_name)
      for condition_name, condition_value in conditions.items():
        query = query.where(
          filter=FieldFilter(condition_name, '==', condition_value)
        )
      results = [self.entity(**result.to_dict()) for result in query.stream()]
      return results
    except Exception:
      return []

  def _list(self):
    results = self.client.collection(self.collection_name).stream()
    entities = [self.entity(**result.to_dict()) for result in results]
    return entities

  def _update(self, task_id: str, update_dict: dict[str, Any]):
    if doc := self.client.collection(self.collection_name).document(task_id):
      doc.update(update_dict)

  def _delete(self, task_id: str) -> None:
    self.client.collection(self.collection_name).document(task_id).delete()
